s3v4_signature_fields = (
    "success_action_status",
    "Content-Disposition",
    "x-amz-meta-user-id",
    "x-amz-meta-config-name",
    "Content-Type",
    "key",
    "x-amz-security-token",
    "x-amz-algorithm",
    "x-amz-credential",
    "x-amz-date",
    "policy",
    "x-amz-signature",
)
